@@ -3,48 +3,54 @@
 'Address'
 from string import Template
 
-from sql import Null
-from sql.conditionals import Case, Coalesce
-from sql.operators import Concat
+from sql import Literal
+from sql.conditionals import Coalesce
+from sql.functions import Substring
+from sql.operators import Concat, Equal
 
-from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
-    sequence_ordered
+from trytond.i18n import gettext
+from trytond.model import (
+    ModelView, ModelSQL, MatchMixin, DeactivableMixin, fields,
+    sequence_ordered, Exclude)
+from trytond.model.exceptions import AccessError
 from trytond.pyson import Eval, If
-from trytond import backend
 from trytond.pool import Pool
+from trytond.rpc import RPC
 from trytond.transaction import Transaction
 from trytond.cache import Cache
-
-__all__ = ['Address', 'AddressFormat']
-
-STATES = {
-    'readonly': ~Eval('active'),
-    }
-DEPENDS = ['active']
+from .exceptions import InvalidFormat
 
 
-class Address(sequence_ordered(), ModelSQL, ModelView):
+class Address(DeactivableMixin, sequence_ordered(), ModelSQL, ModelView):
     "Address"
     __name__ = 'party.address'
     party = fields.Many2One('party.party', 'Party', required=True,
         ondelete='CASCADE', select=True, states={
-            'readonly': If(~Eval('active'), True, Eval('id', 0) > 0),
+            'readonly': Eval('id', 0) > 0,
             },
-        depends=['active', 'id'])
-    name = fields.Char("Building Name", states=STATES, depends=DEPENDS)
-    street = fields.Text("Street", states=STATES, depends=DEPENDS)
-    zip = fields.Char('Zip', states=STATES, depends=DEPENDS)
-    city = fields.Char('City', states=STATES, depends=DEPENDS)
-    country = fields.Many2One('country.country', 'Country',
-        states=STATES, depends=DEPENDS)
+        depends=['id'])
+    party_name = fields.Char(
+        "Party Name",
+        help="If filled, replace the name of the party for address formatting")
+    name = fields.Char("Building Name")
+    street = fields.Text("Street")
+    postal_code = fields.Char("Postal Code")
+    city = fields.Char("City")
+    country = fields.Many2One('country.country', "Country")
+    subdivision_types = fields.Function(
+        fields.MultiSelection(
+            'get_subdivision_types', "Subdivision Types"),
+        'on_change_with_subdivision_types')
     subdivision = fields.Many2One("country.subdivision",
-        'Subdivision', domain=[
+        'Subdivision',
+        domain=[
             ('country', '=', Eval('country', -1)),
-            ('parent', '=', None),
+            If(Eval('subdivision_types', []),
+                ('type', 'in', Eval('subdivision_types', [])),
+                ()
+                ),
             ],
-        states=STATES, depends=['active', 'country'])
-    active = fields.Boolean('Active',
-        help="Uncheck to exclude the address from future use.")
+        depends=['country', 'subdivision_types'])
     full_address = fields.Function(fields.Text('Full Address'),
             'get_full_address')
 
@@ -52,21 +58,23 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
     def __setup__(cls):
         super(Address, cls).__setup__()
         cls._order.insert(0, ('party', 'ASC'))
-        cls._error_messages.update({
-                'write_party': 'You can not modify the party of address "%s".',
-                })
+        cls.__rpc__.update(
+            autocomplete_postal_code=RPC(instantiate=0, cache=dict(days=1)),
+            autocomplete_city=RPC(instantiate=0, cache=dict(days=1)),
+            )
 
     @classmethod
     def __register__(cls, module_name):
-        TableHandler = backend.get('TableHandler')
         cursor = Transaction().connection.cursor()
-        table = TableHandler(cls, module_name)
         sql_table = cls.__table__()
+        table = cls.__table_handler__(module_name)
+
+        # Migration from 5.8: rename zip to postal code
+        table.column_rename('zip', 'postal_code')
 
         super(Address, cls).__register__(module_name)
 
-        # Migration from 2.4: drop required on sequence
-        table.not_null_action('sequence', action='remove')
+        table = cls.__table_handler__(module_name)
 
         # Migration from 4.0: remove streetbis
         if table.column_exist('streetbis'):
@@ -78,12 +86,9 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
                     [value]))
             table.drop_column('streetbis')
 
-    @staticmethod
-    def default_active():
-        return True
-
     _autocomplete_limit = 100
 
+    @fields.depends('country', 'subdivision')
     def _autocomplete_domain(self):
         domain = []
         if self.country:
@@ -98,25 +103,25 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
 
     def _autocomplete_search(self, domain, name):
         pool = Pool()
-        Zip = pool.get('country.zip')
+        PostalCode = pool.get('country.postal_code')
         if domain:
-            records = Zip.search(domain, limit=self._autocomplete_limit)
+            records = PostalCode.search(domain, limit=self._autocomplete_limit)
             if len(records) < self._autocomplete_limit:
                 return sorted({getattr(z, name) for z in records})
         return []
 
-    @fields.depends('city', 'country', 'subdivision')
-    def autocomplete_zip(self):
+    @fields.depends('city', methods=['_autocomplete_domain'])
+    def autocomplete_postal_code(self):
         domain = self._autocomplete_domain()
         if self.city:
             domain.append(('city', 'ilike', '%%%s%%' % self.city))
-        return self._autocomplete_search(domain, 'zip')
+        return self._autocomplete_search(domain, 'postal_code')
 
-    @fields.depends('zip', 'country', 'subdivision')
+    @fields.depends('postal_code', methods=['_autocomplete_domain'])
     def autocomplete_city(self):
         domain = self._autocomplete_domain()
-        if self.zip:
-            domain.append(('zip', 'ilike', '%s%%' % self.zip))
+        if self.postal_code:
+            domain.append(('postal_code', 'ilike', '%s%%' % self.postal_code))
         return self._autocomplete_search(domain, 'city')
 
     def get_full_address(self, name):
@@ -136,9 +141,10 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
                 subdivision_code = subdivision_code.split('-', 1)[1]
         substitutions = {
             'party_name': '',
+            'attn': '',
             'name': getattr(self, 'name', None) or '',
             'street': getattr(self, 'street', None) or '',
-            'zip': getattr(self, 'zip', None) or '',
+            'postal_code': getattr(self, 'postal_code', None) or '',
             'city': getattr(self, 'city', None) or '',
             'subdivision': (self.subdivision.name
                 if getattr(self, 'subdivision', None) else ''),
@@ -148,18 +154,46 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
             'country_code': (self.country.code or ''
                 if getattr(self, 'country', None) else ''),
             }
+        # Keep zip for backward compatibility
+        substitutions['zip'] = substitutions['postal_code']
         if context.get('address_from_country') == getattr(self, 'country', ''):
             substitutions['country'] = ''
         if context.get('address_with_party', False):
-            substitutions['party_name'] = (self.party.full_name
-                if getattr(self, 'party', None) else '')
-        for key, value in substitutions.items():
+            substitutions['party_name'] = self.party_full_name
+        if context.get('address_attention_party', False):
+            substitutions['attn'] = (
+                context['address_attention_party'].full_name)
+        for key, value in list(substitutions.items()):
             substitutions[key.upper()] = value.upper()
         return substitutions
 
+    @property
+    def party_full_name(self):
+        name = ''
+        if self.party_name:
+            name = self.party_name
+        elif self.party:
+            name = self.party.full_name
+        return name
+
     def get_rec_name(self, name):
+        party = self.party_full_name
+        if self.street:
+            street = self.street.splitlines()[0]
+        else:
+            street = None
+        if self.country:
+            country = self.country.code
+        else:
+            country = None
         return ', '.join(
-            filter(None, [self.party.rec_name, self.zip, self.city]))
+            filter(None, [
+                    party,
+                    self.name,
+                    street,
+                    self.postal_code,
+                    self.city,
+                    country]))
 
     @classmethod
     def search_rec_name(cls, name, clause):
@@ -168,9 +202,12 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
         else:
             bool_op = 'OR'
         return [bool_op,
-            ('zip',) + tuple(clause[1:]),
-            ('city',) + tuple(clause[1:]),
             ('party',) + tuple(clause[1:]),
+            ('name',) + tuple(clause[1:]),
+            ('street',) + tuple(clause[1:]),
+            ('postal_code',) + tuple(clause[1:]),
+            ('city',) + tuple(clause[1:]),
+            ('country',) + tuple(clause[1:]),
             ]
 
     @classmethod
@@ -180,8 +217,9 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
             if 'party' in values:
                 for address in addresses:
                     if address.party.id != values['party']:
-                        cls.raise_user_error(
-                            'write_party', (address.rec_name,))
+                        raise AccessError(
+                            gettext('party.msg_address_change_party',
+                                address=address.rec_name))
         super(Address, cls).write(*args)
 
     @fields.depends('subdivision', 'country')
@@ -190,20 +228,32 @@ class Address(sequence_ordered(), ModelSQL, ModelView):
                 and self.subdivision.country != self.country):
             self.subdivision = None
 
+    @classmethod
+    def get_subdivision_types(cls):
+        pool = Pool()
+        Subdivision = pool.get('country.subdivision')
+        selection = Subdivision.fields_get(['type'])['type']['selection']
+        return [(k, v) for k, v in selection if k is not None]
 
-class AddressFormat(MatchMixin, ModelSQL, ModelView):
+    @fields.depends('country')
+    def on_change_with_subdivision_types(self, name=None):
+        pool = Pool()
+        Types = pool.get('party.address.subdivision_type')
+        return Types.get_types(self.country)
+
+
+class AddressFormat(DeactivableMixin, MatchMixin, ModelSQL, ModelView):
     "Address Format"
     __name__ = 'party.address.format'
-    country = fields.Many2One('country.country', "Country")
-    language = fields.Many2One('ir.lang', "Language")
-    active = fields.Boolean("Active",
-        help="Uncheck to exclude the format from future use.")
+    country_code = fields.Char("Country Code", size=2)
+    language_code = fields.Char("Language Code", size=2)
     format_ = fields.Text("Format", required=True,
         help="Available variables (also in upper case):\n"
         "- ${party_name}\n"
         "- ${name}\n"
+        "- ${attn}\n"
         "- ${street}\n"
-        "- ${zip}\n"
+        "- ${postal_code}\n"
         "- ${city}\n"
         "- ${subdivision}\n"
         "- ${subdivision_code}\n"
@@ -215,31 +265,52 @@ class AddressFormat(MatchMixin, ModelSQL, ModelView):
     @classmethod
     def __setup__(cls):
         super(AddressFormat, cls).__setup__()
-        cls._order.insert(0, ('country', 'ASC'))
-        cls._order.insert(1, ('language', 'ASC'))
-        cls._error_messages.update({
-                'invalid_format': ('Invalid format "%(format)s" '
-                    'with exception "%(exception)s".'),
-                })
+        cls._order.insert(0, ('country_code', 'ASC NULLS LAST'))
+        cls._order.insert(1, ('language_code', 'ASC NULLS LAST'))
 
     @classmethod
-    def default_active(cls):
-        return True
+    def __register__(cls, module_name):
+        pool = Pool()
+        Country = pool.get('country.country')
+        Language = pool.get('ir.lang')
+        country = Country.__table__()
+        language = Language.__table__()
+        table = cls.__table__()
+        cursor = Transaction().connection.cursor()
+
+        super().__register__(module_name)
+
+        table_h = cls.__table_handler__()
+
+        # Migration from 5.2: replace country by country_code
+        if table_h.column_exist('country'):
+            query = table.update(
+                [table.country_code],
+                country.select(
+                    country.code,
+                    where=country.id == table.country))
+            cursor.execute(*query)
+            table_h.drop_column('country')
+
+        # Migration from 5.2: replace language by language_code
+        if table_h.column_exist('language'):
+            query = table.update(
+                [table.language_code],
+                language.select(
+                    Substring(language.code, 0, 2),
+                    where=language.id == table.language))
+            cursor.execute(*query)
+            table_h.drop_column('language')
 
     @classmethod
     def default_format_(cls):
         return """${party_name}
 ${name}
 ${street}
-${zip} ${city}
+${postal_code} ${city}
 ${subdivision}
 ${COUNTRY}"""
 
-    @staticmethod
-    def order_language(tables):
-        table, _ = tables[None]
-        return [Case((table.language == Null, 1), else_=0), table.language]
-
     @classmethod
     def create(cls, *args, **kwargs):
         records = super(AddressFormat, cls).create(*args, **kwargs)
@@ -269,34 +340,22 @@ ${COUNTRY}"""
         try:
             Template(self.format_).substitute(
                 **address._get_address_substitutions())
-        except Exception, exception:
-            self.raise_user_error('invalid_format', {
-                    'format': self.format_,
-                    'exception': exception,
-                    })
+        except Exception as exception:
+            raise InvalidFormat(gettext('party.invalid_format',
+                    format=self.format_,
+                    exception=exception)) from exception
 
     @classmethod
     def get_format(cls, address, pattern=None):
-        pool = Pool()
-        Language = pool.get('ir.lang')
-
         if pattern is None:
             pattern = {}
         else:
             pattern = pattern.copy()
         pattern.setdefault(
-            'country', address.country.id if address.country else None)
+            'country_code', address.country.code if address.country else None)
+        pattern.setdefault('language_code', Transaction().language[:2])
 
-        languages = Language.search([
-                ('code', '=', Transaction().language),
-                ], limit=1)
-        if languages:
-            language, = languages
-        else:
-            language = None
-        pattern.setdefault('language', language.id if language else None)
-
-        key = tuple(sorted(pattern.iteritems()))
+        key = tuple(sorted(pattern.items()))
         format_ = cls._get_format_cache.get(key)
         if format_ is not None:
             return format_
@@ -310,3 +369,65 @@ ${COUNTRY}"""
 
         cls._get_format_cache.set(key, format_)
         return format_
+
+
+class SubdivisionType(DeactivableMixin, ModelSQL, ModelView):
+    "Address Subdivision Type"
+    __name__ = 'party.address.subdivision_type'
+    country_code = fields.Char("Country Code", size=2, required=True)
+    types = fields.MultiSelection('get_subdivision_types', "Subdivision Types")
+    _get_types_cache = Cache('party.address.subdivision_type.get_types')
+
+    @classmethod
+    def __setup__(cls):
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_constraints = [
+            ('country_code_unique',
+                Exclude(t, (t.country_code, Equal),
+                    where=t.active == Literal(True)),
+                'party.msg_address_subdivision_country_code_unique')
+            ]
+        cls._order.insert(0, ('country_code', 'ASC NULLS LAST'))
+
+    @classmethod
+    def get_subdivision_types(cls):
+        pool = Pool()
+        Subdivision = pool.get('country.subdivision')
+        selection = Subdivision.fields_get(['type'])['type']['selection']
+        return [(k, v) for k, v in selection if k is not None]
+
+    @classmethod
+    def create(cls, *args, **kwargs):
+        records = super().create(*args, **kwargs)
+        cls._get_types_cache.clear()
+        return records
+
+    @classmethod
+    def write(cls, *args, **kwargs):
+        super().write(*args, **kwargs)
+        cls._get_types_cache.clear()
+
+    @classmethod
+    def delete(cls, *args, **kwargs):
+        super().delete(*args, **kwargs)
+        cls._get_types_cache.clear()
+
+    @classmethod
+    def get_types(cls, country):
+        key = country.code if country else None
+        types = cls._get_types_cache.get(key)
+        if types is not None:
+            return list(types)
+
+        records = cls.search([
+                ('country_code', '=', country.code if country else None),
+                ])
+        if records:
+            record, = records
+            types = record.types
+        else:
+            types = []
+
+        cls._get_types_cache.set(key, types)
+        return types
